# Reference
# Dobot-Communication-Protocol-V1.1.5.pdf
# PDF published on 2020.04.23

# install the pyserial package: pip install pyserial
import serial
from serial.tools import list_ports
# https://docs.python.org/3/library/struct.html
import struct
import threading
import time

class dobotMagician:
    JUMP_XYZ = 0      # Jump, Cartesian coordinate, MOVJ mode, jump above destination by Height
    MOVJ_XYZ = 1      # Cartesian coordinate, each joint moves independently, higher speed
    MOVL_XYZ = 2      # Cartesian coordinate, rectilinear movement
    JUMP_ANGLE = 3    # Jump, Joint coordinate
    MOVJ_ANGLE = 4    # Joint coordinate
    MOVL_ANGLE = 5    # Joint coordinate
    MOVJ_INC = 6      # Angle increment in Joint coordinate system
    MOVL_INC = 7      # Cartesian coordinate increment in Joint coordinate system
    MOVJ_XYZ_INC = 8  # Cartesian coordinate increment in Cartesian coordinate system 
    JUMP_MOVL_XYZ = 9 # Jump, Cartesian coordinate increment in Cartesian coordinate system

    # default values
    _resetHomeParams = (250,0,50,0)
    _resetPTPJointParams = (200.0, 200.0, 200.0, 200.0, 200.0, 200.0, 200.0, 200.0)
    _resetPTPCoordinateParams = (200.0, 100.0, 200.0, 100.0)
    _resetPTPJumpParams = (20.0, 100.0)
    _resetPTPCommonParams = (100.0, 100.0)
    
    currentHomeParams = list(_resetHomeParams)
    currentPTPJointParams = list(_resetPTPJointParams)
    currentPTPCoordinateParams = list(_resetPTPCoordinateParams)
    currentPTPJumpParams = list(_resetPTPJumpParams)
    currentPTPCommonParams = list(_resetPTPCommonParams)
    
    def __init__(self, debugMode=False):
        self._debugMode=debugMode
        self.comPort=""
        self._lock = threading.Lock()
        dobotAlarm=AlarmDescriptions()
        self.alarmState=dobotAlarm.descriptions

    def connect(self, serialPort=''):
        # scan for the first Dobot if serialPort=''
        scanPorts=[]
        for p in list_ports.comports():
            scanPorts.append(p.device)

        if self._debugMode:
            print('Number of available serial ports:',len(scanPorts))
            print('Available ports:',scanPorts)

        if len(scanPorts)==0:
            return "Error: no Com port available."
        
        if serialPort!='':
            if serialPort in scanPorts:
                # remove other ports, leave the specified port only
                scanPorts.clear()
                scanPorts.append(serialPort)
            else:
                return "Error: specified port "+serialPort+" does not exist."

        # scanPorts stores the list of serial ports to verify
        self._ser=serial.Serial()
        self._ser.baudrate=115200
        self._ser.bytesize=serial.EIGHTBITS
        self._ser.parity=serial.PARITY_NONE
        self._ser.stopbits=serial.STOPBITS_ONE
        self._ser.timeout=1 # avoid waiting indefinitely for response due to COM port or Dobot error
        for p in scanPorts:
            if self._debugMode:
                print("Try to find Dobot at",p)
            self.comPort=p
            self._ser.port=p
            self._ser.close()  # To close the port in case it is already opened to avoid open already opened port error
            time.sleep(1) # give Dobot some time to response

            try:
                self._ser.open()
            except serial.SerialException:
                return "Error: "+p+" detected but cannot be opened."
            time.sleep(1) # give Dobot some time to response

            if self._isDobotMagician():
                if self._debugMode:
                    print('Dobot Magician found at',p)
                return "" # successful
            if self._debugMode:
                print(p,' tried but it is not Dobot Magician.')
            self.disconnect()
        return "Error: Dobot Magician cannot be found."

    def disconnect(self):
        # once disconnected, all commands in the queue are not executed.  Dobot stops at once.
        self._lock.acquire() # make sure that other commands/messages Tx/Rx completed
        self._ser.close()
        if self._debugMode:
            print('%s closed.' % self.comPort)
        self._lock.release()

    def _sendCommand(self, DMcommandBytes,wait=False,remarks="",expectedReturnLen=0):
        # return the Dobot response, empty for error (bytes object)
        self._lock.acquire()

        if self._debugMode:
            print("IN waiting:",self._ser.in_waiting,"Out waiting:",self._ser.out_waiting)
            self._printProtocol('TX >> ', DMcommandBytes,remarks)
        self._ser.reset_input_buffer()  # discard all contents in the input buffer (due to error from last command)
        #self._ser.reset_output_buffer() # discard all contents in the output buffer
        self._ser.write(DMcommandBytes)
        
        #s = self._ser.read_all()
        if self._debugMode:
            time.sleep(0.1) # wait for the Dobot to response
            print("IN waiting:",self._ser.in_waiting,"Out waiting:",self._ser.out_waiting)
        if expectedReturnLen>0:        
            s = self._ser.read(expectedReturnLen)
        else:
            s = self._ser.read(100) # for unspecified return length, wait for 100 bytes within timeout limit
        # the data type of returned s = bytes
        
        checksum=0
        for i in s[3:]:
            checksum+=i
        checksum=checksum%256

        error = not( checksum==0 and len(s)>=6 and s[2]+4==len(s))
        if self._debugMode:
            self._printProtocol('RX << ',s,"error" if error else "ok")

        self._lock.release()

        # return empty bytes for error (bytes object)
        if error:
            return bytes([])
        
        if not wait: # return the Dobot response (bytes object)
            return s

        # unpack 4 bytes only (offset 5) to a long integer
        queuedIndex = struct.unpack_from('L',s,5)[0]

        # wait for the current command to be executed (finished)
        while True:
            currentIndex=self.getQueuedCmdCurrentIndex()
            if self._debugMode:
                print("queued index=",queuedIndex,"current index=",currentIndex)
            if queuedIndex<=currentIndex:
                break
            time.sleep(0.2)
        return s # return the Dobot response (bytes object)
        
    def _printProtocol(self,h,DMcommandBytes,remarks=""):
        print(h,end="")
        print(" ".join("%02X" % a for a in DMcommandBytes),"" if remarks=="" else " ("+remarks+")",sep="")

    def _isDobotMagician(self):
        DMcommand=DMprotocol()
        DMcommand.name="Get device name"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=1
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        # the length of the device name is unknown
        DMcommand.expectedReturnLen=0
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            trial+=1
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            if len(response)>0:
                return True
            if trial>=6: # try a maximum of six times, the Dobot sometimes have very poor response time, up to few seconds
                return False
            time.sleep(0.2)
        return False

    def reset(self):
        # clear command queue and reset the parameters
        self.setQueuedCmdStopExec()
        self.setQueuedCmdClear()
        self.setQueuedCmdStartExec()
        self.setEndEffectorSuctionCup(False,False,False)
        self.setEndEffectorLaser(False,False,False)
        self.setHomePosition(self.currentHomeParams[0],self.currentHomeParams[1],\
                             self.currentHomeParams[2],self.currentHomeParams[3])
        self.setPTPJointParams(self.currentPTPJointParams[0],self.currentPTPJointParams[1],\
                               self.currentPTPJointParams[2],self.currentPTPJointParams[3],\
                               self.currentPTPJointParams[4],self.currentPTPJointParams[5],\
                               self.currentPTPJointParams[6],self.currentPTPJointParams[7])
        self.setPTPCoordinateParams(self.currentPTPCoordinateParams[0],self.currentPTPCoordinateParams[1],\
                                    self.currentPTPCoordinateParams[2],self.currentPTPCoordinateParams[3])
        self.setPTPJumpParams(self.currentPTPJumpParams[0],self.currentPTPJumpParams[1])
        self.setPTPCommonParams(self.currentPTPCommonParams[0],self.currentPTPCommonParams[1])
        return self.getAlarmState()

    def getPose(self):
        # return a tuple -- x, y, z, r and j1, j2, j3, j4 (all floating point)
        # r is measured against the x-axis (polar axis)
        # j4 is measured against the forearm  (polar axis)
        # r = j1+j4
        # empty tuple for error
        DMcommand=DMprotocol()
        DMcommand.name="Get Pose"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=10
        DMcommand.ctrl=0
        DMcommand.params=bytearray([])
        DMcommand.expectedReturnLen=38
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            x=struct.unpack_from('f',response,5)[0]
            y=struct.unpack_from('f',response,9)[0]
            z=struct.unpack_from('f',response,13)[0]
            r=struct.unpack_from('f',response,17)[0]
            j1=struct.unpack_from('f',response,21)[0]
            j2=struct.unpack_from('f',response,25)[0]
            j3=struct.unpack_from('f',response,29)[0]
            j4=struct.unpack_from('f',response,33)[0]
            if self._debugMode:
                print('      Current pose =',"x={:.2f} y={:.2f} z={:.2f} r={:.2f} j1={:.2f} j2={:.2f} j3={:.2f} j4={:.2f}"\
                      .format(x,y,z,r,j1,j2,j3,j4))
            return (x,y,z,r,j1,j2,j3,j4)
        else:
            if self._debugMode:
                print("      Get pose error")
            return ()

    def getAlarmState(self):
        # return a list of bytes containing the error codes
        # [-1] means "Get Alarms State error", cannot get the error codes
        # [] means no alarm
        DMcommand=DMprotocol()
        DMcommand.name="Get Alarm State"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=20
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=22
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            alarmBytes=response[5:21]
            errorCodes=[]
            for i in range(16):
                b=alarmBytes[i]
                for j in range(8):
                    if b & 1 ==1:
                        errorCodes.append(i*8+j)
                    b=b>>1
            if self._debugMode:
                for k in errorCodes:
                    print("0x%x"%k,end=" ")
                print()
            return errorCodes
        else:
            if self._debugMode:
                print("      Get Alarm State error")
            return [-1]

    def clearAlarmState(self):
        DMcommand=DMprotocol()
        DMcommand.name="Clear alarm state"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=21
        DMcommand.ctrl=1
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)

    def setHomePosition(self,xHome=250,yHome=0,zHome=50,rHome=0,wait=False):
        # home position x, y, z, r (4 floating point values)
        DMcommand=DMprotocol()
        DMcommand.name="Set home position"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=30
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([])
        # struct.pack returns a bytes object (4 bytes) containing the float format of the xHome
        DMcommand.params.extend(bytearray(struct.pack('f', xHome)))
        DMcommand.params.extend(bytearray(struct.pack('f', yHome)))
        DMcommand.params.extend(bytearray(struct.pack('f', zHome)))
        DMcommand.params.extend(bytearray(struct.pack('f', rHome)))
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        s=self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)
        if len(s)>0:
            self.currentHomeParams=[xHome,yHome,zHome,rHome]
        return s

    # This must be the first movement command to set up the coordinate system.
    def setToHome(self):
        DMcommand=DMprotocol()
        DMcommand.name="Set to home"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=31
        DMcommand.ctrl=3
        DMcommand.params=bytes([0,0,0,0]) # reserved for future use
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,True,DMcommand.name,DMcommand.expectedReturnLen)

    def setEndEffectorLaser(self,ctrl=False,laser=False,wait=False):
        # SW4, power interface of EIO2 (laser) 12V 1A output
        # ctrl=True to activate the SW4 so that EIO2 (laser) voltage = 12V if laser=True
        DMcommand=DMprotocol()
        DMcommand.name="Set End Effector Laser"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=61
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([0,0])
        DMcommand.params[0]=1 if ctrl else 0
        DMcommand.params[1]=1 if laser else 0
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)

    def setEndEffectorSuctionCup(self,ctrl=False,suck=False,wait=False):
        # SW1, power interface of EIO16 (air pump) 12V 1A output
        # ctrl=True to activate the SW1 so that EIO16 (suck) voltage = 12V if suck=True
        DMcommand=DMprotocol()
        DMcommand.name="Set End Effector Suction Cup"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=62
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([0,0])
        DMcommand.params[0]=1 if ctrl else 0
        DMcommand.params[1]=1 if suck else 0
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)

    # Point to point movement
    def getPTPJointParams(self):
        # return a tuple -- velocity and acceleration for each joint (all floating point)
        # empty tuple for error
        DMcommand=DMprotocol()
        DMcommand.name="Get PTP Joint params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=80
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=38
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            v1=struct.unpack_from('f',response,5)[0]
            v2=struct.unpack_from('f',response,9)[0]
            v3=struct.unpack_from('f',response,13)[0]
            v4=struct.unpack_from('f',response,17)[0]
            a1=struct.unpack_from('f',response,21)[0]
            a2=struct.unpack_from('f',response,25)[0]
            a3=struct.unpack_from('f',response,29)[0]
            a4=struct.unpack_from('f',response,33)[0]
            if self._debugMode:
                print('      PTP Joint params =',"v1={:.2f} v2={:.2f} v3={:.2f} v4={:.2f} a1={:.2f} a2={:.2f} a3={:.2f} a4={:.2f}"\
                      .format(v1,v2,v3,v4,a1,a2,a3,a4))
            return (v1,v2,v3,v4,a1,a2,a3,a4)
        else:
            if self._debugMode:
                print("      Get PTP Joint params error")
            return ()

    def setPTPJointParams(self,v1,v2,v3,v4,a1,a2,a3,a4,wait=False):
        DMcommand=DMprotocol()
        DMcommand.name="Set PTP Joint Params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=80
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([])
        DMcommand.params.extend(bytearray(struct.pack('f', v1)))
        DMcommand.params.extend(bytearray(struct.pack('f', v2)))
        DMcommand.params.extend(bytearray(struct.pack('f', v3)))
        DMcommand.params.extend(bytearray(struct.pack('f', v4)))
        DMcommand.params.extend(bytearray(struct.pack('f', a1)))
        DMcommand.params.extend(bytearray(struct.pack('f', a2)))
        DMcommand.params.extend(bytearray(struct.pack('f', a3)))
        DMcommand.params.extend(bytearray(struct.pack('f', a4)))
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        s=self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)
        if len(s)>0:
            self.currentPTPJointParams=[v1,v2,v3,v4,a1,a2,a3,a4]
            return s
        else:
            return []

    def getPTPCoordinateParams(self):
        # return a tuple -- velocity and acceleration for xyz and r (all floating point)
        # empty tuple for error
        DMcommand=DMprotocol()
        DMcommand.name="Get PTP Cartesian params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=81
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=22
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            vxyz=struct.unpack_from('f',response,5)[0]
            vr=struct.unpack_from('f',response,9)[0]
            axyz=struct.unpack_from('f',response,13)[0]
            ar=struct.unpack_from('f',response,17)[0]
            if self._debugMode:
                print('      PTP Cartesian params =',"vxyz={:.2f} vr={:.2f} axyz={:.2f} ar={:.2f}"\
                      .format(vxyz,vr,axyz,ar))
            return (vxyz,vr,axyz,ar)
        else:
            if self._debugMode:
                print("      Get PTP Cartesian params error")
            return ()

    def setPTPCoordinateParams(self,vxyz,vr,axyz,ar,wait=False):
        DMcommand=DMprotocol()
        DMcommand.name="Set PTP Coordinate Params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=81
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([])
        DMcommand.params.extend(bytearray(struct.pack('f', vxyz)))
        DMcommand.params.extend(bytearray(struct.pack('f', vr)))
        DMcommand.params.extend(bytearray(struct.pack('f', axyz)))
        DMcommand.params.extend(bytearray(struct.pack('f', ar)))
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        s=self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)
        if len(s)>0:
            self.currentPTPCoordinateParams=[vxyz,vr,axyz,ar]
            return s
        else:
            return []

    def getPTPJumpParams(self):
        # return a tuple -- lifting height and height limit (all floating point)
        # empty tuple for error
        DMcommand=DMprotocol()
        DMcommand.name="Get PTP Jump params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=82
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            jumpHeight=struct.unpack_from('f',response,5)[0]
            zLimit=struct.unpack_from('f',response,9)[0]
            if self._debugMode:
                print('      PTP Jump params =',"jumpHeight={:.2f} zLimit={:.2f}"\
                      .format(jumpHeight,zLimit))
            return (jumpHeight,zLimit)
        else:
            if self._debugMode:
                print("      Get PTP Jump params error")
            return ()

    def setPTPJumpParams(self,jumpHeight,zLimit,wait=False):
        # jump from A to B
        # if B is above A, jumpHeight is above B
        # if B is below A, jumpHeight is above A
        # max.height = jumpHeight + max(A,B)
        # zLimit is the absolute maximum z-coordinate of the jump movement
        DMcommand=DMprotocol()
        DMcommand.name="Set PTP Jump Params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=82
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([])
        DMcommand.params.extend(bytearray(struct.pack('f', jumpHeight)))
        DMcommand.params.extend(bytearray(struct.pack('f', zLimit)))
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        s=self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)
        if len(s)>0:
            self.currentPTPJumpParams=[jumpHeight,zLimit]
            return s
        else:
            return []

    def getPTPCommonParams(self):
        # return a tuple --  velocity ratio and acceleration ratio (all floating point)
        # empty tuple for error
        DMcommand=DMprotocol()
        DMcommand.name="Get PTP Common params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=83
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=5:
                break
        if len(response)>0:
            velocityRatio=struct.unpack_from('f',response,5)[0]
            accelerationRatio=struct.unpack_from('f',response,9)[0]
            if self._debugMode:
                print('      PTP Common params =',"velocityRatio={:.2f} accelerationRatio={:.2f}"\
                      .format(velocityRatio,accelerationRatio))
            return (velocityRatio,accelerationRatio)
        else:
            if self._debugMode:
                print("      Get PTP Common params error")
            return ()

    def setPTPCommonParams(self,velocityRatio,accelerationRatio,wait=False):
        DMcommand=DMprotocol()
        DMcommand.name="Set PTP Common Params"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=83
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([])
        DMcommand.params.extend(bytearray(struct.pack('f', velocityRatio)))
        DMcommand.params.extend(bytearray(struct.pack('f', accelerationRatio)))
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        s=self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)
        if len(s)>0:
            self.currentPTPCommonParams=[velocityRatio,accelerationRatio]
            return s
        else:
            return []

    # Point to point movement
    def setPTPCmd(self,x,y,z,r,mode=JUMP_XYZ,wait=True):
        DMcommand=DMprotocol()
        DMcommand.name="Set PTP command"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=84
        # use queue
        DMcommand.ctrl=3
        DMcommand.params=bytearray([])
        DMcommand.params.extend([mode])
        DMcommand.params.extend(bytearray(struct.pack('f', x)))
        DMcommand.params.extend(bytearray(struct.pack('f', y)))
        DMcommand.params.extend(bytearray(struct.pack('f', z)))
        DMcommand.params.extend(bytearray(struct.pack('f', r)))
        # use queue, return the current command index
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)

    # Arc movement
    # from current point A, pass through intermediate point B to end point C.
    # A, B and C cannot be on a straight line
    # if A and C are opposite points on a diameter, drawing a semi-circle is not guaranteed
    def setArcCmd(self,intermediateX,intermediateY,intermediateZ,intermediateR,endX,endY,endZ,endR,wait=True):
        DMcommand=DMprotocol()
        DMcommand.name="Set Arc command"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=101
        # use queue
        DMcommand.ctrl=3
        DMcommand.params=bytearray([])
        DMcommand.params.extend(bytearray(struct.pack('f', intermediateX)))
        DMcommand.params.extend(bytearray(struct.pack('f', intermediateY)))
        DMcommand.params.extend(bytearray(struct.pack('f', intermediateZ)))
        DMcommand.params.extend(bytearray(struct.pack('f', intermediateR)))
        DMcommand.params.extend(bytearray(struct.pack('f', endX)))
        DMcommand.params.extend(bytearray(struct.pack('f', endY)))
        DMcommand.params.extend(bytearray(struct.pack('f', endZ)))
        DMcommand.params.extend(bytearray(struct.pack('f', endR)))
        # use queue, return the current command index
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)

    # Set I/O multiplexing
    def setIOMultiplexing(self,IOaddress=0,multiplex=0,wait=False):
        # IOaddress range (1-20)
        # multiplex=1=IOFunctionDO (output)
        # multiplex=2=IOFunctionPWM (output)
        # multiplex=3=IOFunctionDI (input)
        # multiplex=4=IOFunctionADC (input)
        # multiplex=5=IOFunctionPullUpInput (input)
        # multiplex=6=IOFunctionPullDownInput (input)
        if IOaddress<=0 or IOaddress>20:
            return []
        DMcommand=DMprotocol()
        DMcommand.name="Set IO multiplexing"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=130
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([0,0])
        DMcommand.params[0]=IOaddress
        DMcommand.params[1]=multiplex
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)

    # Turn on/off the output pin
    def setIODO(self,IOaddress=0,level=0,wait=False):
        # IOaddress range (1-20), level=1 (high level), level=0 (low level)
        if IOaddress<=0 or IOaddress>20:
            return []
        DMcommand=DMprotocol()
        DMcommand.name="Set IO output"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=131
        DMcommand.ctrl=3 if wait else 1
        DMcommand.params=bytearray([0,0])
        DMcommand.params[0]=IOaddress
        DMcommand.params[1]=level
        DMcommand.expectedReturnLen=14 if wait else 6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,wait,DMcommand.name,DMcommand.expectedReturnLen)


    def setQueuedCmdStartExec(self):
        DMcommand=DMprotocol()
        DMcommand.name="Set QueuedCmdStartExec"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=240
        DMcommand.ctrl=1
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)

    def setQueuedCmdStopExec(self):
        DMcommand=DMprotocol()
        DMcommand.name="Set QueuedCmdStopExec"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=241
        DMcommand.ctrl=1
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)

    def setQueuedCmdClear(self):
        DMcommand=DMprotocol()
        DMcommand.name="Set QueuedCmdClear"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=245
        DMcommand.ctrl=1
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=6
        DMcommandBytes=DMcommand._toBytes()
        return self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)

    def getQueuedCmdCurrentIndex(self):
        # some Dobot models do not reset the queue index to 0
        DMcommand=DMprotocol()
        DMcommand.name="Get QueuedCmdCurrentIndex"
        DMcommand.header=bytes([0xAA,0xAA])
        DMcommand.id=246
        DMcommand.ctrl=0
        DMcommand.params=bytes([])
        DMcommand.expectedReturnLen=14
        DMcommandBytes=DMcommand._toBytes()
        trial=0
        while True:
            response=self._sendCommand(DMcommandBytes,False,DMcommand.name,DMcommand.expectedReturnLen)
            trial+=1
            if len(response)>0 or trial>=10:
                break
        if len(response)>0:
            # can only handle 32 bits, not 64 bits provided by the Dobot
            currentIndex = struct.unpack_from('L',response,5)[0]
            if self._debugMode:
                print('      Current command queue index =',currentIndex)
            return currentIndex
        else:
            if self._debugMode:
                print("      Get command queue index error")
            return 0

#----------------------------------------------------------------------------

    # additional methods for easier control
    # use names from Blocky programming of Dobot Studio

    def setDebug(self,debug=False):
        self._debugMode=debug
        
    def moveToHome(self,mode=JUMP_XYZ,wait=True):
        return self.setPTPCmd(self.currentHomeParams[0],self.currentHomeParams[1],\
                              self.currentHomeParams[2],self.currentHomeParams[3],mode,wait)

    def getAllStatus(self):
        # return a dictionary
        allStatus= {
            "Com. port":self.comPort,
            "PTP Joint Params":tuple(self.currentPTPJointParams),
            "PTP Coordinate Params":tuple(self.currentPTPCoordinateParams),
            "PTP Jump Params":tuple(self.currentPTPJumpParams),
            "PTP Common Params":tuple(self.currentPTPCommonParams),
            "Home position":tuple(self.currentHomeParams),
            "Current position":self.getPose(),
            "Queue index":self.getQueuedCmdCurrentIndex()
        }
        return allStatus


    def home(self):
        self.setToHome()

    def setMotionRatio(self,vRatio=100.0,aRatio=100.0):
        self.setPTPCommonParams(vRatio,aRatio,False)

    def setJointSpeed(self,v=200.0,a=200.0):
        self.setPTPJointParams(v,v,v,v,a,a,a,a,False)

    def setCoordinateSpeed(self,v=200.0,a=200.0):
        params=self.getPTPCoordinateParams
        if len(params)>0:
            self.setPTPCoordinateParams(v,params(1),a,params(3),False)

    def setJumpHeight(self,h=20.0):
        self.setPTPJumpParams(h,100.0,False)

    def jumpTo(self,x,y,z):
        pose=self.getPose()
        self.setPTPCmd(x,y,z,pose[3],self.JUMP_XYZ,True)

    def moveTo(self,x,y,z):
        # absolute Cartesian coordinates, rectilinear movement
        # the end-effector is rotated so that the direction r with reference to the x-axis is not changed
        pose=self.getPose()
        self.setPTPCmd(x,y,z,pose[3],self.MOVL_XYZ,True)
        
    def moveDistance(self,x,y,z):
        # relative cartesian position, r kept unchanged
        # relative movement
        self.setPTPCmd(x,y,z,0,self.MOVJ_XYZ_INC,True)

    def setR(self,r):
        # r is measured against the x-axis (polar axis)
        # only rotate the end-effector
        self.setPTPCmd(0,0,0,r,self.MOVJ_XYZ_INC,True)  # relative cartesian position, x,y,z kept unchanged
        
    def setJointAngle(self,j1,j2,j3):
        # j4 is measured against the forearm  (polar axis)
        # j4 is not changed
        # the end-effector is not rotated so that the direction r is changed
        pose=self.getPose()
        self.setPTPCmd(j1,j2,j3,pose[7],self.MOVJ_ANGLE,True)  # Joint coordinate, each joint moves independently
        
    def getCurrentCoordinate(x,y,z,r):
        # not tested/confirmed
        pose=self.getPose()
        return (pose[0],pose[1],pose[2],pose[3])  # Cartesian coordinates only
        
    def getJointAngle(j1,j2,j3,j4):
        # not tested/confirmed
        pose=self.getPose()
        return (pose[4],pose[5],pose[6],pose[7])  # Joint coordinates only
        
    def set5VOutput(self,address,state=0):
        # EIO 10, 13
        # state=0 (off), state=1 (on)
        self.setIODO(address,state,False)

    def set12VOutput(self,address,state=0):
        # EIO 2, 3, 16, 17
        # e.g. forearm SW4 EIO2 12V output, address=2
        # state=0 (off), state=1 (on)
        self.setIODO(address,state,False)

    def suctionCup(self,state=0):
        # state=0 (off), state=1 (on)
        if state==1:
            setEndEffectorSuctionCup(True,True,False)
        else:
            setEndEffectorSuctionCup(False,False,False)
        
#----------------------------------------------------------------------------
    
class DMprotocol:
    # DMprotocol is a bytes object, a list of bytes
    def __init__(self):
        self.header=bytes([0xAA,0xAA])
        self.len=0
        self.id=0
        # bit1(0=immediate, 1=queue), bit0(0=get/read, 1=set/write)
        # 0 immediate read -- all read commands
        # 1 immediate set
        # 2 queue read
        # 3 queue set
        self.ctrl=0
        self.params=bytes([])
        self.checksum=0
        self.expectedReturnLen=0

    # update the len and checksum, convert the protocol object to bytearray for printing(debug) and sending
    def _toBytes(self):
        sum=self.id+self.ctrl
        for i in self.params:
            sum += i
        sum=sum%256
        checksum=(256-sum)%256
        #update the len and the checksum
        self.len=2+len(self.params)
        self.checksum=checksum
        b=bytearray(self.header)
        b.append(self.len)
        b.append(self.id)
        b.append(self.ctrl)
        b.extend(self.params)
        b.append(self.checksum)
        return b

class AlarmDescriptions:
    def __init__(self):
        self.descriptions={
            0:"After resetting system",
            1:"Receive undefined instruction",
            2:"File system error",
            3:"Failed communication between MCU and FPGA when system is initializing",
            4:"Get an error value of angle sensor",
            16:"The target point is in abnormal position in Cartesian coordinate system",
            17:"The target point is out of the workspace, that causes inverse resolve alarm",
            18:"The inverse resolve of target is out of the limitation",
            19:"There are some repetitive points in ARC or JUMP_MOVEL mode",
            20:"Arc command error, invalid points",
            21:"Set a wrong parameter in JUMP mode",
            32:"Motion trajectory is in singularity position that causes inverse resolve alarm in Cartesian coordinate system",
            33:"Motion is out of workspace that causes inverse resolve alarm when Dobot Magician moves",
            34:"The motion is out of the limitation when Dobot Maigican moves",
            48:"Joint 1 is overspeed",
            49:"Joint 2 is overspeed",
            50:"Joint 3 is overspeed",
            51:"Joint 4 is overspeed",
            64:"Joint 1 moves to the positive limitation area",
            65:"Joint 1 moves to the negative limitation area",
            66:"Joint 2 moves to the positive limitation area",
            67:"Joint 2 moves to the negative limitation area",
            68:"Joint 3 moves to the positive limitation area",
            69:"Joint 3 moves to the negative limitation area",
            70:"Joint 4 moves to the positive limitation area",
            71:"Joint 4 moved to the negative limitation area",
            72:"Parallelogram is stretched to the positive limitation area",
            73:"Parallelogram is stretched to the negative limitation area",
            80:"Joint 1 loses step",
            81:"Joint 2 loses step",
            82:"Joint 3 loses step",
            83:"Joint 4 loses step"
        }

